from __future__ import annotations
from functools import lru_cache
from typing import Dict, List

from .member import Member
from .client import client
from .message import Message
from .util import Waiter, fetch_cursored


class Channel:
    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        creator: Member,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.creator = creator

    @property
    def mention(self) -> str:
        return f"<#{self.id}>"

    def invite(self, members: List[Member]):
        Waiter(tire=3)
        client().conversations_invite(
            channel=self.id, users=[member.id for member in members]
        )

        Message.post(
            self,
            f"""{len(members)}人を{self.name}に招待しました
招待者リスト: {", ".join([member.name for member in members])}
""",
        )

    @property
    def members(self) -> List[Member]:
        response = fetch_cursored(
            lambda cursor: client().conversations_members(
                cursor=cursor, channel=self.id
            ),
            "members",
            tire=4,
        )
        return [Member.fetch(member_id) for member_id in response]

    @staticmethod
    @lru_cache()
    def fetch_all():
        response = fetch_cursored(
            lambda cursor: client().conversations_list(
                cursor=cursor, exclude_archived=True
            ),
            "channels",
            tire=2,
        )

        channels = [
            Channel(
                channel["id"],
                channel["name"],
                channel["purpose"]["value"],
                Member.fetch(channel["creator"]),
            )
            for channel in response
        ]
        id_list = [channel.id for channel in channels]

        return dict(zip(id_list, channels))

    @classmethod
    def fetch_public_channel(cls, channel_id: str):
        channels = cls.fetch_all()
        if channel_id in channels.keys():
            return channels[channel_id]

        # キャッシュを削除する
        cls.fetch_all.cache_clear()
        channels = cls.fetch_all()
        if channel_id in channels.keys():
            return channels[channel_id]

        # それでも見つからない場合
        raise RuntimeError(f"Channel id {channel_id} not found")

    @classmethod
    def fetch_by_response(cls, response: Dict[str, str]) -> Channel:
        if response.get("channel_type", "") == "im":
            # DMの場合
            return cls(
                response["channel"],
                response["user"],
                "",
                Member.fetch(response["user"]),
            )
        else:
            return cls.fetch_public_channel(response.get("channel", ""))



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Optional

from .util import Waiter
from .client import client

if TYPE_CHECKING:
    from .channel import Channel


@dataclass
class Message:
    """メッセージを表現するクラス"""

    channel: Channel
    text: str
    timestamp: str
    root_timestamp: str = ""

    def add_reaction(self, reaction_name: str):
        Waiter(tire=3)
        client().reactions_add(
            channel=self.channel.id, name=reaction_name, timestamp=self.timestamp
        )

    def update(self, new_message: str):
        self.text = new_message
        Waiter(tire=3)
        client.chat_update(channel=self.channel.id, ts=self.timestamp, text=self.text)

    @staticmethod
    def post(
        channel: Channel,
        message: str,
        *,
        image_url: Optional[str] = None,
        reply2: Optional[Message] = None,
        reply2channel: bool = False,
    ):
        def _blocks():
            if image_url is None:
                return None
            else:
                return [
                    {
                        "type": "section",
                        "text": {"type": "plain_text", "text": message},
                    },
                    {
                        "type": "image",
                        "image_url": image_url,
                        "alt_text": "image is not loaded",
                    },
                ]

        if reply2 is None:
            reply2 = Message(channel, "", "", "")

        Waiter(tire=3)
        response: Dict = client().chat_postMessage(
            channel=channel.id,
            text=message,
            thread_ts=reply2.timestamp,
            reply_broadcast=reply2channel,
            blocks=_blocks(),
        )  # type: ignore

        return Message(channel, message, response["ts"], reply2.timestamp)



from typing import Optional
from slack_bolt import App
from slack_sdk import WebClient

_app: Optional[App] = None
_client: Optional[WebClient] = None


def raise_uninitialized_error():
    raise RuntimeError("Please use init() to initialize slack2python")


def app() -> App:
    if _app is None:
        raise_uninitialized_error()
    return _app


def client() -> WebClient:
    if _client is None:
        raise_uninitialized_error()
    return _client


def init(bot_user_oauth_token: str, signing_secret: str):
    """initialize slack2python

    To get the arguments of this function, follow the steps.
    1. Go to https://api.slack.com/apps/
    2. Click your app
    3. Copy 'Singing Secret'
    4. Click 'OAuth & Permissions'
    5. Add at least 1 bot token scopes if no scope registered
    6. Copy 'Bot User OAuth Token'

    Args:
        bot_user_oauth_token (str): Bot User OAuth Token
        signing_secret (str): Singing Secret
    """
    global _app, _client

    if _app is not None or _client is not None:
        raise RuntimeError("init() cannot be executed twice or more")

    _app = App(
        token=bot_user_oauth_token,
        signing_secret=signing_secret,
    )
    _client = _app.client


def start(port: int = 8000):
    """start a slack bolt session

    Args:
        port (int): port to listen slack events
    """
    app().start(port=port)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slack2python']

package_data = \
{'': ['*']}

install_requires = \
['slack-bolt>=1.16.1,<2.0.0']

setup_kwargs = {
    'name': 'slack2python',
    'version': '0.1.1',
    'description': 'a python wrapper of slack bolt',
    'long_description': '# slack2python\na python wrapper of slack bolt\n',
    'author': '2lu3',
    'author_email': 'hi2lu3@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from functools import lru_cache
from typing import Dict

from .client import client
from .util import Waiter, fetch_cursored


class Member:
    """メンバー1人を表現するクラス
    __init__(), fetch()の2つの初期化方法がある
    """

    def __init__(
        self,
        id: str,
        name: str,
    ):
        self.id = id
        self.name = name

    @property
    def mention(self) -> str:
        """slack中でメンションする場合このメソッドが返す文字列を使う必要がある"""
        return f"<@{self.id}>"

    @staticmethod
    @lru_cache()
    def fetch_all() -> Dict[str, Member]:
        """ワークスペースに存在するアカウントをすべて問い合わせる
        botや解約済みアカウントなどは省く
        """

        def _is_valid_member(member: Dict) -> bool:
            """そのmemberがbotなのかユーザーなのか判定する"""

            def _has_prohibited_keys(member: Dict):
                """botや解約済みのプロパティを持っているか判定する"""
                # 以下の属性が1つでもtrueだと省きたい
                prohibited_keys = [
                    "is_bot",
                    "deleted",
                    "is_restricted",
                    "is_ultra_restricted",
                ]
                return sum([member.get(key, 0) for key in prohibited_keys]) != 0

            def _is_slackbot(member: Dict):
                """slackbotかどうか判定する
                slackbotは _has_prohibited_keywards に該当しないため"""

                return member["id"] == "USLACKBOT"

            return not _has_prohibited_keys(member) and not _is_slackbot(member)

        members = fetch_cursored(
            lambda cursor: client().users_list(cursor=cursor, exclude_arrived=True),
            "members",
            2,
        )
        members = [
            Member(member["id"], member["profile"]["real_name"])
            for member in members
            if _is_valid_member(member)
        ]
        keys = [member.id for member in members]
        return dict(zip(keys, members))

    @classmethod
    @lru_cache
    def fetch(cls, member_id: str) -> Member:
        """メンバーidからメンバーを返す"""
        # return fetch_all()'s member to avoid ask slack server frequently
        if member_id in cls.fetch_all().keys():
            return cls.fetch_all()[member_id]

        # キャッシュを削除して、slack serverに問い合わせる
        cls.fetch_all.cache_clear()
        if member_id in cls.fetch_all().keys():
            return cls.fetch_all()[member_id]

        # それでもない場合、botなどの可能性があるので個別に問い合わせる
        Waiter(tire=4)
        response: Dict = client().users_info(user=member_id)  # type: ignore
        if response["ok"] == False and response["error"] != "user_not_found":
            raise ValueError(f"Invalid member id: {member_id}")
        else:
            return Member(
                response["user"]["id"], response["user"]["profile"]["real_name"]
            )

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)



from datetime import datetime, timedelta
from time import sleep
from typing import Callable, Dict, List
from slack_sdk.web import SlackResponse


class Waiter:
    """SlackのRateLimitを回避するために一定時間待つクラス"""

    last_executed_time: Dict[int, datetime] = dict(
        [(i, datetime.now()) for i in range(1, 5)]
    )
    time_required: Dict[int, timedelta] = {
        1: timedelta(seconds=60 / 1),
        2: timedelta(seconds=60 / 20),
        3: timedelta(seconds=60 / 50),
        4: timedelta(seconds=60 / 100),
    }

    def __init__(self, *, tire: int):
        """tireごとに必要な時間までsleepする

        Args:
            tire (int): tire. please read https://api.slack.com/docs/rate-limits
        """
        if self.required_seconds(tire=tire) > 0:
            Waiter.last_executed_time[tire] = datetime.now() + timedelta(
                seconds=self.required_seconds(tire=tire)
            )
            sleep(self.required_seconds(tire=tire))
        else:
            Waiter.last_executed_time[tire] = datetime.now()

    @classmethod
    def required_seconds(cls, *, tire: int) -> float:
        elapsed_time = datetime.now() - Waiter.last_executed_time[tire]
        required_time = Waiter.time_required[tire] - elapsed_time

        if required_time.total_seconds() < 0:
            return 0
        else:
            return required_time.total_seconds()


def fetch_cursored(
    fetcher: Callable[[str], SlackResponse], concat_name: str, tire: int
) -> List:
    """next_cursorを使って分割ダウンロードを行う

    Args:
        fetcher (Callable[[str], SlackResponse]): next_cursorを引数として、Slack Web APIから受けとったSlackResponseを返すような関数
        concat_name (str): 分割してダウンロードしたレスポンスを1つにまとめるときのキー
        tire (int): ダウンロードするAPIのTire

    Returns:
        List: ダウンロードした結果を1つにまとめて返す
    """
    result: List = []
    next_cursor = ""
    while True:
        Waiter(tire=tire)

        response: Dict = fetcher(next_cursor)  # type: ignore
        next_cursor = response["response_metadata"]["next_cursor"]
        result.extend(response[concat_name])

        if next_cursor == "":
            break

    return result
